/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_CHAN_LIST_HH
#define SENDS_CHAN_LIST_HH
#include "types.hh"
#include "epoch.hh"
#include "chan_type.hh"
#include "data_type.hh"
#include "sortedvector.hh"

#include <string>
#include <iosfwd>

namespace sends {

    /**  The channel list entry contains the channel name, data type, 
      *  access code and a list of frame types that include the channel.
      *  The channel name and channel type are both needed to provide a 
      *  unique identification of the channel because in some cases, the
      *  same channel name is used for both raw and decimated (RDS) data 
      *  or for minute and second trends. The frame types are used to 
      *  construct the frame names, e.g. "<site>-<type>-<gps>-<dt>.gwf".
      *  The initial implementation (2008.01.22) assumes that the frame
      *  lists do not change in time. It may become necessary to add a 
      *  valid time range to each source data entry to reliably guide the
      *  channel source selection.
      *  @brief Channel list entry
      *  @author John Zweizig
      *  @version 1.2; Last modified 2010.05.12
      *  @ingroup sends_base
      */
    class chan_entry {
    public:

	/**  The source entry class contains a source name and the time epoch
	  *  during which the channel was included in the specified source.
	  *  \brief Channel source list entry.
	  */
	class source_entry {
	public:
	    /**  Constant value to represent the end of all time
	     */
	    static const gps_type inf_gps=1999999999;
	public:
	    /**  Empty source entry constructor
	     */
	    source_entry(void);

	    /**  Source entry data constructor
	     */
	    source_entry(const std::string& name, const epoch& ep);

	    /**  Source entry destructor
	     */
	    ~source_entry(void);

	    /**  Get the end of the epoch for which the channel is included
	      *  in this source type.
	      *  \brief GPS of end of epoch.
	      *  \return Epoch end time.
	      */
	    gps_type end_gps(void) const;

	    /**  Get the %epoch for which the channel is included
	      *  in this source type.
	      *  \brief Reference to epoch.
	      *  \return Epoch constant reference.
	      */
	    const epoch& refEpoch(void) const;

	    /**  Get the source name.
	      *  \brief Source name.
	      *  \return constant reference to source name string.
	      */
	    const std::string& name(void) const;

	    /**  Merge time of specified entries.
	      *  \brief merge epochs.
	      *  \param se Entry to be merged with this.
	      */
	    void merge(const source_entry& se);

	    /**  Test whether entries correspond to the same source in
	      *  overlapping or abutting times.
	      *  \param se Source entry to be tested for overlap with this
	      *  \return Boolean true of entries overlap.
	      */
	    bool overlap(const source_entry& se) const;

	    /**  Compare two entries
	      *  \brief Compare entries.
	      *  \param se Entry to be compared.
	      *  \return True if this should precede specified entry in 
	      *          ordered list.
	      */
	    bool operator<(const source_entry& se) const;

	    /**  Get the start of the epoch for which the channels is included
	      *  in this source type.
	      *  \brief GPS of start of epoch.
	      *  \return Epoch start time.
	      */
	    gps_type start_gps(void) const;

	    /**  Swap the contents of this instance with that of the argument
	      *  %source_entry.
	      *  \brief Swap contents with another %source_entry.
	      *  \param x Source entry with which to swap contents.
	      */
	    void swap(source_entry& x);

	    /**  Test whether the named source is valid during the
	      *  specified %epoch.
	      *  \brief Test source validity.
	      *  \param t Test gps %epoch.
	      *  \return True if source is valid during the specified %epoch.
	      */
	    bool test_epoch(const epoch& t) const;

	    /**  Test whether the named source is valid at the specified 
	      *  gps time.
	      *  \brief Test source validity.
	      *  \param t Test gps time.
	      *  \return True if source entry valid at specified time.
	      */
	    bool test_gps(gps_type t) const;
	private:
	    ///  Source name string.
	    std::string _name;
	    ///  Validity epoch
	    epoch    _epoch;
	};

	/**  Source vector.
	  */
	typedef sorted_vector<source_entry> source_vect;

	/**  Source vector iterator.
	  */
	typedef source_vect::iterator       source_iter;

	/**  Source vector constant iterator.
	  */
	typedef source_vect::const_iterator const_source_iter;

	/**  Data type used to represent the number of entries in the source 
	  *  list.
	  *  @brief Source count data type.
	  */
	typedef index_type source_count;

    public:
	/**  Construct a channel entry.
	  *  @brief Channel constructor
	  *  @param name   Full channel name.
	  *  @param type   Type of channel
	  *  @param access Minimum level required to access this channel.
	  *  @param dtype  Data type code.
	  *  @param rate   Sample rate.
	  *  @param source Type code of frame containing the channel.
	  */
	chan_entry(const std::string& name, chan_type type, long access=0, 
		   data_type dtype=data_type::kReal4, double rate=0,
		   const std::string& source="", const epoch& ep=epoch(0));

	/**  Add a new source frame-type string. The source type string
	  *  contains the site identifier and the frame type, separated
	  *  by a hyphen, e.g. "H-R" and start and end times. If the 
	  *  specified epoch overlaps or abuts an existing epoch for the 
	  *  same source, the epochs are merged.
	  *  @brief Add a frame source
	  *  @param ftype Source type string.
	  *  @param start Start of source validity epoch.
	  *  @param end   End of source validity epoch.
	  */
	void addSource(const std::string& ftype, const epoch& ep);

	/**  Test whether the channel type is available at the specified time.
	  *  @brief Test for source availability.
	  *  @param t Test time.
	  *  @return True if available.
	  */
	bool available(gps_type t) const;

	/**  Get the access level required to reference this channel.
	  *  @brief Get channel access level.
	  *  @return Access level required for this channel.
	  */
	long getAccess(void) const {return mAccess;}

	/**  Get the data type code.
	  *  @brief Data type code.
	  *  @return The data type code.
	  */
	data_type dtype(void) const {return mDType;}

	/**  The data length in bytes for the specified time interval.
	  *  @brief Data length.
	  *  @return Data length for the specified delta-t.
	  */
	index_type data_length(double dt) const;

	/**  Get the length of a data word in bytes.
	  *  @brief Data work length.
	  *  @return Data word length in bytes.
	  */
	int data_size(void) const {return mDType.size();}

	/**  Get a constant reference to the channel name string.
	  *  @brief Get channel name.
	  *  @return const reference to string containing the channel name.
	  */
	const std::string& name(void) const {return mName;}

	/**  Get the number of source frame-types.
	  *  @brief Number of sources.
	  *  @return Number of entries in the source list.
	  */
	source_count getNSource(void) const;

	/**  Get a constant reference to the i'th source frame type string.
	  *  @brief Get source string
	  *  @param inx Index into the source list.
	  *  @return Constant reference to string containing the source type.
	  */
	const std::string& getSource(source_count inx=0) const;

	/**  Get the specified source descriptor
	  *  @brief Number of sources.
	  *  @return Number of entries in the source list.
	  */
	const source_entry& source(int i) const;

	/**  Get the channel type code.
	  *  @return Channel type code.
	  */
	chan_type type(void) const {return mType;}

	/**  See if channel is available in the named frame type. Note that
	  *  if the variation of the channel content of the frame types with
	  *  time is to be supported, this function will have to add a
	  *  time argument.
	  *  @brief Test whether the channel exists in the specified source.
	  *  @param src Source name to be tested.
	  *  @return True if the channel exists in the specified source.
	  */
	bool is_source(const std::string& src, gps_type t=0) const;

	/**  Merge source list from the argument entry into this entry.
	  *  This method is implemented by looping over the entries in 
	  *  the argument list and using the addSource method to add in 
	  *  sources one at a time. Although this is hideously inefficient,
	  *  I expect that the source list will never become more than 2-3 
	  *  entries long and will only be modified at start-time.
	  *  @brief merge source lists
	  *  @param ce Channel entry containing list to be merged.
	  */
	void merge(const chan_entry& ce);

	/**  Get the memory used by this entry.
	  *  \brief Return the memory use.
	  *  \return Number of bytes of memory used.
	  */
	index_type mem_use(void) const;

	/**  The channel name and type fields are compared between two
	  *  channel entries. The name field is given greater significance
	  *  than the type field. The result is true if this entry name is 
	  *  less <i>i.e.</i> 
	  *  \code name < ent.name || (name == ent.name && type < ent.type)
	     \endcode
	  *  The comparison of the type values is false if either operand has 
	  *  an unknown type.
	  *  @brief Compare two channel entries.
	  *  @param ent Channel entry to be compared with this.
	  *  @return True if this entry is less.
	  */
	bool operator<(const chan_entry& ent)  const;

	/**  The channel name and type fields are compared to between two
	  *  channel entries and the result is true the entries are equal.
	  *  @brief Compare two channel entries.
	  *  @param ent Channel entry to be compared with this.
	  *  @return True if the entries are equal.
	  */
	bool operator==(const chan_entry& ent) const;

	/**  The channel name and type fields are compared to between two
	  *  channel entries and the result is true if the entries are not 
	  *  equal.
	  *  @brief Compare two channel entries.
	  *  @param ent Channel entry to be compared with this.
	  *  @return True if the entries are not equal.
	  */
	bool operator!=(const chan_entry& ent) const;

	/**  Return the channel sample rate.
	  *  @brief Sample rate.
	  *  @return Sample rate in Hz.
	  */
	double rate(void) const;

	/**  Set the access level required to see this channel.
	  *  @brief Set the access level.
	  *  @param acclvl Access level of user requesting access.
	  */
	void set_access(long acclvl);

	/**  Swap two channel entries.
	  *  @brief Swap channel entries.
	  *  @param ce Reference to the channel entry with which this 
	  *            channel is to be swapped.
	  */
	void swap(chan_entry& ce);

	/**  Test whether access to the channel is allowed to a user with
	  *  the specified access level.
	  *  @brief Test if access is allowed
	  *  @param acclvl Access level of user requesting access.
	  *  @return True if channel access is allowed to specified level user.
	  */
	bool test_access(long acclvl) const;

	/**  Write a one line channel description to the specified output
	  *  stream.
	  *  \brief Write a channel description.
	  *  \param out   Standard output stream reference.
	  *  \param times Include valid time range.
	  *  \return Output stream reference.
	  */
	std::ostream& write(std::ostream& out, bool times=false) const;

    private:
	/// Channel name
	std::string mName;

	/// Channel type
	chan_type   mType;

	///  Access mask
	long        mAccess;

	///  Data type
	data_type   mDType;

	/// Sample rate
	double      mRate;

	/// List of source frame types
	source_vect mSource;
    };

    /**  Print entry name
      */
    std::ostream& operator<<(std::ostream& out, const chan_entry& ent);

    /**  Master channel list.
      *  @brief Channel list.
      *  @author John Zweizig
      *  @version 1.0; Last modified: January 23, 2008
      *  @ingroup sends_base
      */
    class chan_list {
    public:
	/**  Construct an empty channel list. The list_available flag is 
	  *  cleared to prevent clients from using the channel list information.
	  *  @brief Construct empty channel list.
	  *  @param ndef Number of entries to preallocate.
	  */
	chan_list(unsigned long ndef=100000);

	/**  Add the specified channel to the channel list. If the list is 
	  *  available for read access, an exception is thrown.
	  *  @brief Add a channel to the channel list.
	  *  @param ce Channel entry
	  */
	chan_index add_channel(const chan_entry& ce);

	/** Clear the list and reset the available flag.
	  */
	void clear(void);

	/**  Dump the contents of the channel list to the specified output
	  *  stream.
	  *  @brief Dump the channel list contents.
	  *  @param out Output stream.
	  *  \param times Include valid time ranges.
	  *  @return Reference to the argument output stream.
	  */
	std::ostream& dump(std::ostream& out, bool times=false) const;

	/** Enable the list for read access.
	  */
	void enable(void);

	/**  Find the specified channel entry in the list. A runtime_error
	  *  exception is thrown if the channel is not found.
	  *  @brief Find channel entry
	  *  @param name Name of channel to be found.
	  *  @param t    Channel type.
	  *  @exception runtime_error if not found.
	  *  @return Index into the channel list.
	  */
	chan_index find(const std::string& name, chan_type t) const;

	/**  Find the specified channel entry in the list. A runtime_error
	  *  exception is thrown if the channel is not found.
	  *  @brief Find channel entry
	  *  @exception runtime_error if not found.
	  *  @param ent Channel entry (name, type) to be found.
	  *  @exception runtime_error if not found.
	  *  @return Index into the channel list.
	  */
	chan_index find(const chan_entry& ent) const;

	/**  Find the channel that best matches the specified name, 
	  *  channel type, rate and gps time. An exception is thrown
	  *  if the specified channel name is not in the list, or if 
	  *  the listed channel with the given name do not match the 
	  *  other parameters.
	  *  @brief Find channel with the nearest parameters
	  *  @param name Name of channel to be found.
	  *  @param t    Channel type.
	  *  @param rate Sample rate
	  *  @param gps  GPS time at which the channel is needed (or 0).
	  *  @exception runtime_error if not found.
	  *  @return index into the channel list.
	  */
	chan_index find_best(const std::string& name, chan_type t,
			     double rate, gps_type gps) const;

	/**  Find the lower bound index for the specified channel name.
	  *  @brief Find the lower bound index.
	  *  @param name Name of channel to be found.
	  *  @return index into the channel list.
	  */
	chan_index lower_bound(const std::string& name, 
			       chan_type t=chan_type::kUnknown) const;

	/**  Get the memory used by this list.
	  *  \brief Return the memory use.
	  *  \return Number of bytes of memory used.
	  */
	index_type mem_use(void) const;

	/**  Fill the channel list from a file of channel definitions. The
	  *  file contains one channel definition per line. Comments may be
	  *  included by following any definition text with a pound sign and 
	  *  the comment text. A channel definition may be continued on one 
	  *  or more additional lines by typing a backslash as the last 
	  *  character of the non-terminal line(s). definition fields are 
	  *  separated by white-space characters an contain the following
	  *  <table>
	  *  <tr><td>Column</td><td>Contents</td></tr>
	  *  <tr><td>1</td><td>Channel name</td></tr>
	  *  <tr><td>2</td><td>Channel type</td></tr>
	  *  <tr><td>3</td><td>Minimum access level</td></tr>
	  *  <tr><td>4</td><td>Data type</td></tr>
	  *  <tr><td>5</td><td>Sample rate</td></tr>
	  *  <tr><td>6 - n</td><td>Source frame(s)</td></tr>
	  *  </table>
	  *  @brief Read channel file.
	  *  @param file  Path to a file of channel definitions.
	  *  @param start Start of validity epoch
	  *  @param end   End of validity epoch.
	  */
	void read(const std::string& file, const epoch& ep=epoch(0));

	/**  Extend the channel list to contain at least the specified number
	  *  of entries. No action is taken if the current capacity is at least
	  *  the specified number of entries.
	  *  @brief reserve the specified number of entries
	  *  @param nMax new capacity
	  */
	void reserve(chan_index nMax);

	/**  Return the number of active entries in the channel list.
	  *  @returns The number of entries in the channel list.
	  */
	chan_index size(void) const;

	/**  Find the upper bound index for the specified channel name.
	  *  @brief Find the lower bound index.
	  *  @param name Name of channel to be found.
	  *  @return index into the channel list.
	  */
	chan_index upper_bound(const std::string& name, 
			       chan_type t=chan_type::kUnknown) const;

	/**  Get a constant reference to the entry specified by the channel
	  *  index.
	  *  @brief Reference the specified entry.
	  *  @param inx Index to the desired channel entry.
	  *  @exception std::out_of_range is thrown if the index is invalid.
	  *  @returns constant reference to the channel entry.
	  */
	const chan_entry& operator[](chan_index inx) const;

	/**  Get a reference to the entry specified by the channel index.
	  *  @brief Reference the specified entry.
	  *  @param inx Index to the desired channel entry.
	  *  @exception std::out_of_range is thrown if the index is invalid.
	  *  @returns Reference to the channel entry.
	  */
	chan_entry& operator[](chan_index inx);

    public:
	///  Channel vector data type
	typedef sorted_vector<chan_entry> chan_vect;
	///  Channel vector iterator
	typedef chan_vect::iterator       chan_iter;
	/// Constant channel vector iterator
	typedef chan_vect::const_iterator const_chan_iter;

	/**  Merge an unsorted chan_entry vector into the channel list.
	 */
	void merge(const chan_vect::base_type& vec);

    private:
	bool list_available;
	chan_vect mChanList;
    };

    //==================================  Define source_entry inline methods
    inline gps_type
    chan_entry::source_entry::end_gps(void) const {
	return _epoch.end();
    }

    inline const std::string&
    chan_entry::source_entry::name(void) const {
	return _name;
    }

    inline bool 
    chan_entry::source_entry::operator<(const source_entry& se) const {
	return _name < se._name || (_name == se._name && _epoch < se._epoch);
    }

    inline bool 
    chan_entry::source_entry::overlap(const source_entry& se) const {
	return _name == se._name && _epoch.overlap(se._epoch);
    }

    inline const epoch& 
    chan_entry::source_entry::refEpoch(void) const {
	return _epoch;
    }

    inline gps_type
    chan_entry::source_entry::start_gps(void) const {
	return _epoch.start();
    }

    inline bool
    chan_entry::source_entry::test_epoch(const epoch& epic) const {
	return _epoch.overlapx(epic);
    }

    inline bool
    chan_entry::source_entry::test_gps(gps_type t) const {
	return !t || _epoch.contains(t);
    }

    //==================================  Define chan_entry inline methods
    inline index_type  
    chan_entry::data_length(double dt) const {
	return index_type(dt * mRate * data_size() + 0.5);
    }

    inline chan_entry::source_count 
    chan_entry::getNSource(void) const {
	return mSource.size();
    }

    inline const std::string& 
    chan_entry::getSource(source_count inx) const {
	return mSource[inx].name();
    }

    inline bool 
    chan_entry::operator<(const chan_entry& ent) const {
	if (mName < ent.mName)  return true;
	if (ent.mName != mName) return false;
	//------ note that kUnknown is smallest so ent.mType can't be 
	//       kUnknown if mType < ent.mType.
	if (mType != chan_type::kUnknown && mType < ent.mType) return true;
	return false;
    }

    inline bool 
    chan_entry::operator==(const chan_entry& ent) const {
	return (mName == ent.mName && mType == ent.mType);
    }

    inline bool 
    chan_entry::operator!=(const chan_entry& ent) const {
	return (mName != ent.mName || mType != ent.mType);
    }

    inline double
    chan_entry::rate(void) const {
	return mRate;
    }

    inline const chan_entry::source_entry&
    chan_entry::source(int i) const {
	return mSource[i];
    }

    inline bool
    chan_entry::test_access(long acclvl) const {
	return acclvl >= mAccess;
    }

    //==================================  Define chan_list inline methods
    inline chan_index 
    chan_list::find(const chan_entry& ent) const {
	return mChanList.findex(ent);
    }

    inline chan_index 
    chan_list::find(const std::string& name, chan_type type) const{
	return find(chan_entry(name, type));
    }

    inline const chan_entry& 
    chan_list::operator[](chan_index inx) const {
	return mChanList[inx];
    }

    inline chan_entry& 
    chan_list::operator[](chan_index inx) {
	return mChanList[inx];
    }

    inline chan_index
    chan_list::size(void) const {
	return mChanList.size();
    }

    //==================================  Define extern Channel_List
    extern chan_list    Channel_List;

} // namespace sends

#endif // !defined(SENDS_CHAN_LIST_HH)
